/*
	SORTIER.C

	Module zum C-Kurs, Teil 16
	http://members.aol.com/pcdbascht/ckmain.htm

  	(c) 1998, 1999 Sebastian Cyris
*/

#include "struktur.h"
#include "suchen.h"
#include "sortier.h"
#include <stdio.h>
#include <stdlib.h>


/*
	Insertion Sort
	
	Werte werden bei jedem Aufruf in der Liste eingereiht
	bergeben werden Liste und einzureihendes Element
*/

EVL *InsertionSort ( EVL *Liste, double Wert ) 
{
	EVL		*temp = Liste;
	EVL		*neu = malloc(sizeof(EVL));
	neu->Element	= Wert;
	neu->next		= NULL;

	if(Liste == NULL) return neu;
	
	/* Liste durchwandern, notfalls einreihen */
	while (temp->next != NULL)
	{
		if (temp->next->Element > Wert)
		{
	
			neu->next	= temp->next;
			temp->next	= neu;
			return Liste;
		}
		temp = temp->next;
	} 

	/* Am Ende angelangt, Element einreihen */
	temp->next	= neu;

	/* Wenn letztes Element > neues Element, dann vertauschen */
	if (temp->Element > Wert)
	{
		neu->Element	= temp->Element;
		temp->Element	= Wert;
	}

	/* Listenanfang zurckgeben */
	return Liste;
}


/* 
	Bubble Sort

	Es wird bei einer Liste erst das erste und dann alle nachfolgenden
	Elemente mit der Restliste verglichen und mit kleineren Elementen vertauscht
*/

EVL *BubbleSort		( EVL *Liste )
{
	EVL		*Rest, *Akt = Liste;
	double	swap;

	/* Liste ist leer oder hat nur ein Element */
	if ((Liste == NULL)||(Liste->next == NULL)) return Liste;

	/* Sortierverfahren Bubblesort */
	do
	{
		Rest = Akt;
		do 
		{
			Rest = Rest->next;
			/* Wenn der Restwert kleiner als Vergleichswert, dann vertauschen */
			if (Rest->Element < Akt->Element)
			{
				swap			= Rest->Element;
				Rest->Element	= Akt->Element;
				Akt->Element	= swap;
			}
		} while (Rest->next != NULL);
		Akt = Akt->next;

	} while (Akt->next != NULL);

	return Liste;
}

/* 
	Quicksort

	Sortieren durch das "Teile und Herrsche"-Prinzip
	Es werden teilfelder solange in sich selbst aufgeteilt und sortiert
	bis eine vollstndige Sortierung vorliegt
*/

void quicksort (int *sort, int l, int r)  
{ 
	int i, j, Temp; 

	/* solange sich beide Grenzen nicht berhren */ 
	if (r > l) 
	{ 

		i= l - 1; 
		j= r; 

		/* durchlaufen, solange j > i ist */ 
		do 
		{ 

			/* 
			von rechts und links annhern, 
			solange die Werte kleiner als 
			der Bezugswert ist 
			*/ 
			while (sort [++i] < sort [r]); 
			while (sort [--j] > sort [r]); 

			/* Index i und j vertauschen */ 
			Temp = sort [i]; 
			sort [i]= sort [j]; 
			sort [j]= Temp;

		} while (j > i); 

		/* Index i, j und r rotieren */ 
		sort [j]= sort [i]; 
		sort [i]= sort [r]; 
		sort [r]= Temp; 

		/* Teilmengen sortieren */ 
		quicksort (sort, l, i - 1); 
		quicksort (sort, i + 1, r);
	}
}


/*
	sortierter Baum 

	Durch den Einfgemechanismus ergibt sich eine Sortierung auf
	eine spezielle Weise, die erst durch den Wiedergabealgorithmus
	zur entfaltung, d.h. zur Sortierung kommt
*/

Knoten	*SortBaumEinf	(Knoten *Baum, char Wert)
{
	/* neues Knotenelement */
	Knoten	*neu	= malloc (sizeof (Knoten));
	
	neu->key	= Wert;
	neu->l		= NULL;
	neu->r		= NULL;

	/* leerer Baum , Element = Wurzel */
	if (Baum == NULL) return neu;

	/* Wo einreihen ?  */
	if (Wert <= Baum->key) Baum->l = SortBaumEinf (Baum->l , Wert);
	else Baum->r = SortBaumEinf (Baum->r , Wert);
	
	return Baum;
}

void	SortBaumAusgeben	(Knoten *Baum)
{
	if (Baum!=NULL)
	{
		SortBaumAusgeben (Baum->l);
		printf("%c ",Baum->key);
		SortBaumAusgeben  (Baum->r);
	}
}

Knoten *MinElement(Knoten *Baum) 
{ 
	if (Baum == NULL) return Baum;
	while(Baum->l != NULL ) Baum = Baum->l; 
	return Baum; 
}

Knoten *MaxElement(Knoten *Baum) 
{ 
	if (Baum == NULL) return Baum;
	while(Baum->r != NULL ) Baum = Baum->r; 
	return Baum; 
}

/* Heapsort */

void downheap (int *feld, int N, int k) 
{ 
	int j,v; 

	v = feld[k]; 
	while ( k<= N/2 ) 
	{ 
		j = k + k; 

		if (j<N && feld[j]<feld[j+1]) j++; 
		if (v >= feld[j]) break; 
		
		feld[k] = feld[j];  
		k = j; 
	} 

	feld[k]=v; 
}

void heapsort (int *feld, int N)
{
	int k, t;

	for (k = N/2 ; k>=0; k--) downheap ( feld, N, k );
	
	while (N > 0)
	{
		t = feld[0];
		feld[0] = feld[N];
		feld[N] = t;
		downheap (feld, --N, 0);
	}
}