/*
	SUCHEN.C

	Module zum C-Kurs, Teil 17
	http://members.aol.com/pcdbascht/ckmain.htm

  	(c) 1998, 1999 Sebastian Cyris
*/

#include "struktur.h"
#include "suchen.h"
#include "sortier.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
	Es wird ein Feld der Lnge nach verglichen
*/

int SimpelSuchen (int *feld, int Anzahl, int Key) 
{ 
	int Index; 
	for (Index = 0; Index<Anzahl; Index++) 
		if ( feld[Index] == Key ) return Index;

	return -1; 
}

/* binary search */

int binaeresSuchen (int *feld, int Anzahl, int key) 
{ 
	int l=0, r=Anzahl-1, mitte;

	while( r>=l )
	{
		mitte = (l+r)/2;
		if( key == feld[mitte] ) return mitte;
		if( key < feld[mitte] )	r = mitte-1;
		else l = mitte+1;
	}
	return -1;

}

/*
	Es wird ein vorsortierter binrer Baum durchsucht

	Ist das Element gefunden, so wird der entsprechende Knoten
	zurckgegeben, ansonten NULL
*/

Knoten *BaumDurchsuchen (Knoten *Baum, char key)
{
	if (Baum != NULL && Baum->key != key)
	{
		if (Baum->key > key)	return BaumDurchsuchen (Baum->l, key);
		else return BaumDurchsuchen (Baum->r, key);
	}
	return Baum;
}
/*
	Naives Textsuchverfahren

	Es wird der Index zurckgegeben, an dem der Suchstring steht
	bei keiner bereinstimmung wird -1 zurckgegeben
*/

int		NaiveTextSuche	(char *vergleich, char *such)
{
	int laenge_such		= strlen(such) ;
	int	laenge_vergl	= strlen(vergleich) ;
	int i, j;			/* Laufvariablen */

	for ( i=0; i<(laenge_vergl-laenge_such); i++)
	{
		j=0;
		while (vergleich[i+j] == such[j++]) if (j==laenge_such) return i;
	}
	return -1;
}

/*
	Knuth Morris Pratt Verfahren zum Stringsuchen 

	Wenn -1 zurckgegeben wird, wurde der String nicht gefunden,
	ansonsten wird der Index zurckgegeben, an dem der Teilstring liegt
*/

int *KMP_Tabelle( char *gesucht ) 
{ 
	int i,j; /* Hilfsvariable */ 
	int suchlaenge = strlen (gesucht); 
	int *next  = malloc ( suchlaenge * sizeof ( int )); 
	
	/* Speicheranforderung schlug fehl oder kein Feld vorhanden */ 
	if (next == NULL) return NULL; 
	
	next[i=0] = j = -1; 

	 /* next-Tabelle bilden */ 
	do  
	{ 
		if (( j<0 ) || ( gesucht[i] == gesucht[j] )) 
			next[++i] = ++j;
		else 
			j = next[j];
	} while ( gesucht[i] ); 

	return next;
}

int  KnuthMorrisPratt (char *feld, char *gesucht ) 
{ 

	/* Initialisierung der next-Tabelle */ 
	int i,j; /* Hilfsvariable */ 
	int suchlaenge = strlen (gesucht); 
	int feldlaenge = strlen (feld); 
	
	int *next =  KMP_Tabelle( gesucht ) ; /* Nexttabelle */ 

	/* Fehler bei der Speicherbelegung */ 
	if (next == NULL ) return -1; 

	/* Knuth-Morris-Pratt - Algorithmus zur Textsuche*/ 

	j=0; 

	for (i=0; i<feldlaenge; i++) 
	{ 

		/*	
			endlosschleife, die ausgefhrt wird 
			solange Buchstaben sich berlagern 
		*/ 

		  for (;;) 
		  { 

			   /* Wenn mindestens 1 Buchstb. bereinstimmt */ 
			   if (feld[i] == gesucht[j]) 
			   { 
					j++; 

					/* hochzaehlen wieviel Buchstaben schon gleich sind */ 
					/* bei allen den Index zurckgeben, ab welchem Punkt */ 
					if (j == suchlaenge) return i-suchlaenge; 
					break;
			   } 

			   /* keine bereinstimmung, evtl. Sprung ausfhren */ 
			   else if (j == 0 ) break; 
			   else j = next[j];
		  }
	 } 

	 /* sind am Ende, keine bereinstimmung */ 
	 return -1;

}